package dfki.protege.flora_tab;

import java.util.*;
import java.io.*;
import javax.swing.*;


public class Flora {

  ExternalInference flora;
  boolean useStdError = false;
  String floraCommand = null;
  JTextArea dribbleTextArea;

  public Flora(String floraCommand, boolean useStdErr, 
	       JTextArea dribbleTextArea) {
    flora = null;
    this.floraCommand = floraCommand;
    this.useStdError = useStdErr;
    this.dribbleTextArea = dribbleTextArea;
  }

  public void start() throws Exception {
    String[] command = new String[] {floraCommand, "-m", "32000"};
    flora = new ExternalInference(command, useStdError);
    dribble("%%% starting " + floraCommand + "\n\n");
    expect("| ?- ");
    dialogue("[flora].", "| ?- ");
    dialogue("flora_shell.", "flora ?- ");
  }

  public boolean running() {
    return flora != null;
  }

  public void loadKB(String kb) throws Exception {
    dialogue("['" + kb + "'].", "flora ?- ");
  }

  public String query(String query) throws Exception {
    query = query.trim();
    if (!(query.endsWith(".")))
      query = query + ".";
    return dialogue(query, "flora ?- ");
  }

  public void halt() throws Exception {
    tell("halt.");
    // flora.stop(); // stop process ...
    flora = null;
  }


  // utils

  void expect(String expectString) throws Exception {
    dribble(flora.expect(expectString));
  }

  void tell(String string) throws Exception {
    dribble(string);
    flora.tell(string);
  }

  String dialogue(String command, String expectString) throws Exception {
    if (!command.endsWith("\n"))
      command = command + "\n";
    dribble(command);
    return dribble(flora.dialogue(command, expectString));
  }

  String dribble(StringBuffer buffer) {
    return dribble(buffer.toString());
  }

  String dribble(String text) {
    if (dribbleTextArea != null)
      dribbleTextArea.append(text);
    else
      System.out.println(text);
    return text;
  }


}


