package dfki.protege.flora_tab;

import java.awt.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import edu.stanford.smi.protege.action.*;
import edu.stanford.smi.protege.event.*;
import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.util.*;
import edu.stanford.smi.protege.ui.*;

public class InstancesList extends SelectableContainer implements Disposable {

    private SelectableList itsList;
    private Project itsProject;
    private AllowableAction itsDeleteAction;

    private FrameListener itsInstanceFrameListener = new FrameAdapter() {
        public void browserTextChanged(FrameEvent event) {
            super.browserTextChanged(event);
            repaint();
        }
    };

    public InstancesList(Project project) {
        itsProject = project;
        Action viewAction = createViewAction();
        // itsList = ComponentFactory.createSelectableList(viewAction, true);
        itsList = ComponentFactory.createSelectableList(viewAction);
        itsList.setCellRenderer(new FrameRenderer());
        itsList.addListSelectionListener(new ListSelectionListener() {
            public void valueChanged(ListSelectionEvent event) {
                notifySelectionListeners();
            }
        });
        setLayout(new BorderLayout());
        LabeledComponent c = new LabeledComponent("Instances", new JScrollPane(itsList));
        c.addHeaderButton(viewAction);
        // c.addHeaderButton(createReferencersAction());
        // c.addHeaderButton(createDeleteAction());
        c.setFooterComponent(new ListFinder(itsList, "Find Instance"));
        add(c);
        setSelectable(itsList);
    }

    public void dispose() {
        removeListeners();
    }

    private SimpleListModel getModel() {
        return (SimpleListModel) itsList.getModel();
    }

    public void setInstances(Collection instances) {
        removeListeners();
        // Collections.sort(instances, new FrameComparator());
        getModel().setValues(instances);
        addListeners();
        notifySelectionListeners();
    }

    public void initializeSelection() {
        if (itsList.getModel().getSize() > 0) {
            itsList.setSelectedIndex(0);
        }
    }

    private void removeListeners() {
        Iterator i = getModel().getValues().iterator();
        while (i.hasNext()) {
            Instance instance = (Instance) i.next();
            instance.removeFrameListener(itsInstanceFrameListener);
        }
    }

    private void addListeners() {
        Iterator i = getModel().getValues().iterator();
        while (i.hasNext()) {
            Instance instance = (Instance) i.next();
            instance.addFrameListener(itsInstanceFrameListener);
        }
    }

    private Action createViewAction() {
        return new ViewAction("View Instance", this) {
            public void onView(Object o) {
                itsProject.show((Instance) o);
            }
        };
    }

    private Action createDeleteAction() {
        itsDeleteAction = new DeleteInstancesAction("Delete Selected Instances", this);
        return itsDeleteAction;
    }

    /*
    private Action createReferencersAction() {
        return new ReferencersAction(itsProject, this);
    }
    */

    public void onSelectionChange() {
        // Log.enter(this, "onSelectionChange");
        boolean editable = isSelectionEditable();
        ComponentUtilities.setDragAndDropEnabled(itsList, editable);
    }

    private boolean isSelectionEditable() {
        boolean isEditable = true;
        Iterator i = getSelection().iterator();
        while (i.hasNext()) {
            Instance instance = (Instance) i.next();
            if (!instance.isEditable()) {
                isEditable = false;
                break;
            }
        }
        return isEditable;
    }

    public JComponent getDragComponent() {
        return itsList;
    }
}


